from gigachat.pydantic_v1 import BaseModel


class AccessToken(BaseModel):
    """Токен доступа"""

    access_token: str
    """Сгенерированный Access Token"""
    expires_at: int
    """Unix-время завершения действия Access Token в миллисекундах"""



from typing import List

from gigachat.models.model import Model
from gigachat.pydantic_v1 import BaseModel, Field


class Models(BaseModel):
    """Доступные модели"""

    data: List[Model]
    """Массив объектов с данными доступных моделей"""
    object_: str = Field(alias="object")
    """Тип сущности в ответе, например, список"""



from http import HTTPStatus
from typing import Any, AsyncIterator, Dict, Iterator, Optional

import httpx

from gigachat.exceptions import AuthenticationError, ResponseError
from gigachat.models import Chat, ChatCompletionChunk

EVENT_STREAM = "text/event-stream"


def _get_kwargs(
    *,
    chat: Chat,
    access_token: Optional[str] = None,
    client_id: Optional[str] = None,
    session_id: Optional[str] = None,
    request_id: Optional[str] = None,
) -> Dict[str, Any]:
    headers = {
        "Accept": EVENT_STREAM,
        "Cache-Control": "no-store",
    }
    if access_token:
        headers["Authorization"] = f"Bearer {access_token}"
    if client_id:
        headers["X-Client-ID"] = client_id
    if session_id:
        headers["X-Session-ID"] = session_id
    if request_id:
        headers["X-Request-ID"] = request_id

    return {
        "method": "POST",
        "url": "/chat/completions",
        "json": {**chat.dict(exclude_none=True), **{"stream": True}},
        "headers": headers,
    }


def _parse_chunk(line: str) -> Optional[ChatCompletionChunk]:
    name, _, value = line.partition(": ")
    if name == "data":
        if value == "[DONE]":
            return None
        else:
            return ChatCompletionChunk.parse_raw(value)
    else:
        return None


def _check_content_type(response: httpx.Response) -> None:
    content_type, _, _ = response.headers.get("content-type", "").partition(";")
    if content_type != EVENT_STREAM:
        raise httpx.TransportError(f"Expected response Content-Type to be '{EVENT_STREAM}', got {content_type!r}")


def _check_response(response: httpx.Response) -> None:
    if response.status_code == HTTPStatus.OK:
        _check_content_type(response)
    elif response.status_code == HTTPStatus.UNAUTHORIZED:
        raise AuthenticationError(response.url, response.status_code, b"", response.headers)
    else:
        raise ResponseError(response.url, response.status_code, b"", response.headers)


def sync(
    client: httpx.Client,
    *,
    chat: Chat,
    access_token: Optional[str] = None,
    client_id: Optional[str] = None,
    session_id: Optional[str] = None,
    request_id: Optional[str] = None,
) -> Iterator[ChatCompletionChunk]:
    kwargs = _get_kwargs(
        chat=chat, access_token=access_token, client_id=client_id, session_id=session_id, request_id=request_id
    )
    with client.stream(**kwargs) as response:
        _check_response(response)
        for line in response.iter_lines():
            if chunk := _parse_chunk(line):
                yield chunk


async def asyncio(
    client: httpx.AsyncClient,
    *,
    chat: Chat,
    access_token: Optional[str] = None,
    client_id: Optional[str] = None,
    session_id: Optional[str] = None,
    request_id: Optional[str] = None,
) -> AsyncIterator[ChatCompletionChunk]:
    kwargs = _get_kwargs(
        chat=chat, access_token=access_token, client_id=client_id, session_id=session_id, request_id=request_id
    )
    async with client.stream(**kwargs) as response:
        _check_response(response)
        async for line in response.aiter_lines():
            if chunk := _parse_chunk(line):
                yield chunk



from typing import Optional

from gigachat.models.messages import Messages
from gigachat.pydantic_v1 import BaseModel


class Choices(BaseModel):
    """Ответ модели"""

    message: Messages
    """Сгенерированное сообщение"""
    index: int
    """Индекс сообщения в массиве начиная с нуля"""
    finish_reason: Optional[str] = None
    """Причина завершения гипотезы"""



from http import HTTPStatus
from typing import Any, Dict, Optional

import httpx

from gigachat.exceptions import AuthenticationError, ResponseError
from gigachat.models import AccessToken


def _get_kwargs(*, url: str, credentials: str, scope: str, request_id: Optional[str] = None) -> Dict[str, Any]:
    headers = {
        "Authorization": f"Bearer {credentials}",
    }

    if request_id:
        headers["RqUID"] = request_id

    return {
        "method": "POST",
        "url": url,
        "data": {"scope": scope},
        "headers": headers,
    }


def _build_response(response: httpx.Response) -> AccessToken:
    if response.status_code == HTTPStatus.OK:
        return AccessToken(**response.json())
    elif response.status_code == HTTPStatus.UNAUTHORIZED:
        raise AuthenticationError(response.url, response.status_code, response.content, response.headers)
    else:
        raise ResponseError(response.url, response.status_code, response.content, response.headers)


def sync(
    client: httpx.Client, *, url: str, credentials: str, scope: str, request_id: Optional[str] = None
) -> AccessToken:
    kwargs = _get_kwargs(url=url, credentials=credentials, scope=scope, request_id=request_id)
    response = client.request(**kwargs)
    return _build_response(response)


async def asyncio(
    client: httpx.AsyncClient, *, url: str, credentials: str, scope: str, request_id: Optional[str] = None
) -> AccessToken:
    kwargs = _get_kwargs(url=url, credentials=credentials, scope=scope, request_id=request_id)
    response = await client.request(**kwargs)
    return _build_response(response)



from typing import Optional

from gigachat.models.messages_chunk import MessagesChunk
from gigachat.pydantic_v1 import BaseModel


class ChoicesChunk(BaseModel):
    """Ответ модели в потоке"""

    delta: MessagesChunk
    """Короткое сообщение"""
    index: int
    """Индекс сообщения в массиве начиная с нуля"""
    finish_reason: Optional[str] = None
    """Причина завершения гипотезы"""



from typing import List

from gigachat.models.choices_chunk import ChoicesChunk
from gigachat.pydantic_v1 import BaseModel, Field


class ChatCompletionChunk(BaseModel):
    """Ответ модели в потоке"""

    choices: List[ChoicesChunk]
    """Массив ответов модели в потоке"""
    created: int
    """Дата и время создания ответа в формате Unix time"""
    model: str
    """Название модели, которая вернула ответ"""
    object_: str = Field(alias="object")
    """Название вызываемого метода"""



from gigachat.pydantic_v1 import BaseModel


class Token(BaseModel):
    """Токен доступа"""

    tok: str
    """Сгенерированный Access Token"""
    exp: int
    """Unix-время завершения действия Access Token в миллисекундах"""



from http import HTTPStatus
from typing import Any, Dict, Optional

import httpx

from gigachat.exceptions import AuthenticationError, ResponseError
from gigachat.models import Chat, ChatCompletion


def _get_kwargs(
    *,
    chat: Chat,
    access_token: Optional[str] = None,
    client_id: Optional[str] = None,
    session_id: Optional[str] = None,
    request_id: Optional[str] = None,
) -> Dict[str, Any]:
    headers = {}

    if access_token:
        headers["Authorization"] = f"Bearer {access_token}"
    if client_id:
        headers["X-Client-ID"] = client_id
    if session_id:
        headers["X-Session-ID"] = session_id
    if request_id:
        headers["X-Request-ID"] = request_id

    return {
        "method": "POST",
        "url": "/chat/completions",
        "json": chat.dict(exclude_none=True, exclude={"stream"}),
        "headers": headers,
    }


def _build_response(response: httpx.Response) -> ChatCompletion:
    if response.status_code == HTTPStatus.OK:
        return ChatCompletion(**response.json())
    elif response.status_code == HTTPStatus.UNAUTHORIZED:
        raise AuthenticationError(response.url, response.status_code, response.content, response.headers)
    else:
        raise ResponseError(response.url, response.status_code, response.content, response.headers)


def sync(
    client: httpx.Client,
    *,
    chat: Chat,
    access_token: Optional[str] = None,
    client_id: Optional[str] = None,
    session_id: Optional[str] = None,
    request_id: Optional[str] = None,
) -> ChatCompletion:
    kwargs = _get_kwargs(
        chat=chat, access_token=access_token, client_id=client_id, session_id=session_id, request_id=request_id
    )
    response = client.request(**kwargs)
    return _build_response(response)


async def asyncio(
    client: httpx.AsyncClient,
    *,
    chat: Chat,
    access_token: Optional[str] = None,
    client_id: Optional[str] = None,
    session_id: Optional[str] = None,
    request_id: Optional[str] = None,
) -> ChatCompletion:
    kwargs = _get_kwargs(
        chat=chat, access_token=access_token, client_id=client_id, session_id=session_id, request_id=request_id
    )
    response = await client.request(**kwargs)
    return _build_response(response)



from typing import List

from gigachat.models.choices import Choices
from gigachat.models.usage import Usage
from gigachat.pydantic_v1 import BaseModel, Field


class ChatCompletion(BaseModel):
    """Ответ модели"""

    choices: List[Choices]
    """Массив ответов модели"""
    created: int
    """Дата и время создания ответа в формате Unix time"""
    model: str
    """Название модели, которая вернула ответ"""
    usage: Usage
    """Данные об использовании модели"""
    object_: str = Field(alias="object")
    """Название вызываемого метода"""



from typing import List, Optional

from gigachat.models.messages import Messages
from gigachat.pydantic_v1 import BaseModel

LATEST_MODEL = "GigaChat:latest"


class Chat(BaseModel):
    """Параметры запроса"""

    model: str = LATEST_MODEL
    """Название модели, от которой нужно получить ответ"""
    messages: List[Messages]
    """Массив сообщений"""
    temperature: Optional[float] = None
    """Температура выборки в диапазоне от ноля до двух"""
    top_p: Optional[float] = None
    n: Optional[int] = None
    """Количество вариантов ответов, которые нужно сгенерировать для каждого входного сообщения"""
    stream: Optional[bool] = None
    """Указывает, что сообщения надо передавать по частям в потоке"""
    max_tokens: Optional[int] = None
    """Максимальное количество токенов, которые будут использованы для создания ответов"""
    repetition_penalty: Optional[float] = None
    """Количество повторений слов"""
    update_interval: Optional[float] = None
    """Интервал в секундах между отправкой токенов в потоке"""
    profanity_check: Optional[bool] = None



from gigachat.models.access_token import AccessToken
from gigachat.models.chat import Chat
from gigachat.models.chat_completion import ChatCompletion
from gigachat.models.chat_completion_chunk import ChatCompletionChunk
from gigachat.models.choices import Choices
from gigachat.models.choices_chunk import ChoicesChunk
from gigachat.models.messages import Messages
from gigachat.models.messages_chunk import MessagesChunk
from gigachat.models.messages_role import MessagesRole
from gigachat.models.model import Model
from gigachat.models.models import Models
from gigachat.models.token import Token
from gigachat.models.usage import Usage

__all__ = (
    "AccessToken",
    "Chat",
    "ChatCompletion",
    "ChatCompletionChunk",
    "Choices",
    "ChoicesChunk",
    "Messages",
    "MessagesChunk",
    "MessagesRole",
    "Model",
    "Models",
    "Token",
    "Usage",
)



from gigachat.pydantic_v1 import BaseModel


class Usage(BaseModel):
    """Данные об использовании модели"""

    prompt_tokens: int
    """Количество токенов во входящем сообщении"""
    completion_tokens: int
    """Количество токенов, сгенерированных моделью"""
    total_tokens: int
    """Общее количество токенов"""



import logging
from typing import Any, Dict, Optional

from gigachat.pydantic_v1 import BaseSettings, root_validator

ENV_PREFIX = "GIGACHAT_"

# BASE_URL = "https://beta.saluteai.sberdevices.ru/v1"
BASE_URL = "https://gigachat.devices.sberbank.ru/api/v1"
AUTH_URL = "https://ngw.devices.sberbank.ru:9443/api/v2/oauth"
SCOPE = "GIGACHAT_API_CORP"


class Settings(BaseSettings):
    base_url: str = BASE_URL
    """Адрес относительно которого выполняются запросы"""
    auth_url: str = AUTH_URL
    credentials: Optional[str] = None
    scope: str = SCOPE

    access_token: Optional[str] = None
    model: Optional[str] = None
    user: Optional[str] = None
    password: Optional[str] = None

    timeout: float = 30.0
    verify_ssl_certs: bool = True

    use_auth: bool = True
    verbose: bool = False

    class Config:
        env_prefix = ENV_PREFIX

    @root_validator
    def check_credentials(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values["use_auth"]:
            use_secrets = values["credentials"] or values["access_token"] or (values["user"] and values["password"])
            if not use_secrets:
                logging.warning("Please provide GIGACHAT_CREDENTIALS environment variables.")

        return values



from gigachat.models.messages_role import MessagesRole
from gigachat.pydantic_v1 import BaseModel


class Messages(BaseModel):
    """Сообщение"""

    role: MessagesRole
    """Роль автора сообщения"""
    content: str
    """Текст сообщения"""

    class Config:
        use_enum_values = True



import logging
from functools import cached_property
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Iterator,
    Optional,
    TypeVar,
    Union,
)

import httpx

from gigachat.api import get_model, get_models, post_auth, post_chat, post_token, stream_chat
from gigachat.exceptions import AuthenticationError
from gigachat.models import (
    AccessToken,
    Chat,
    ChatCompletion,
    ChatCompletionChunk,
    Model,
    Models,
    Token,
)
from gigachat.settings import Settings

T = TypeVar("T")

logger = logging.getLogger(__name__)


def _get_kwargs(settings: Settings) -> Dict[str, Any]:
    """Настройки для подключения к API GIGACHAT"""
    return {
        "base_url": settings.base_url,
        "verify": settings.verify_ssl_certs,
        "timeout": httpx.Timeout(settings.timeout),
    }


def _get_auth_kwargs(settings: Settings) -> Dict[str, Any]:
    """Настройки для подключения к серверу авторизации OAuth 2.0"""
    return {
        "verify": settings.verify_ssl_certs,
        "timeout": httpx.Timeout(settings.timeout),
    }


def _parse_chat(chat: Union[Chat, Dict[str, Any]], model: Optional[str]) -> Chat:
    payload = Chat.parse_obj(chat)
    if model:
        payload.model = model
    return payload


def _build_access_token(token: Token) -> AccessToken:
    return AccessToken(access_token=token.tok, expires_at=token.exp)


class _BaseClient:
    _access_token: Optional[AccessToken] = None

    def __init__(
        self,
        *,
        base_url: Optional[str] = None,
        auth_url: Optional[str] = None,
        credentials: Optional[str] = None,
        scope: Optional[str] = None,
        access_token: Optional[str] = None,
        model: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        timeout: Optional[float] = None,
        verify_ssl_certs: Optional[bool] = None,
        use_auth: Optional[bool] = None,
        verbose: Optional[bool] = None,
    ) -> None:
        config = {k: v for k, v in locals().items() if k != "self" and v is not None}
        self._settings = Settings(**config)
        if self._settings.access_token:
            self._access_token = AccessToken(access_token=self._settings.access_token, expires_at=0)

    @property
    def token(self) -> Optional[str]:
        if self._settings.use_auth and self._access_token:
            return self._access_token.access_token
        else:
            return None

    @property
    def _use_auth(self) -> bool:
        return self._settings.use_auth

    def _check_validity_token(self) -> bool:
        """Проверить время завершения действия токена"""
        if self._access_token:
            # _check_validity_token
            return True
        return False

    def _reset_token(self) -> None:
        """Сбросить токен"""
        self._access_token = None


class GigaChatSyncClient(_BaseClient):
    """Синхронный клиент GigaChat"""

    @cached_property
    def _client(self) -> httpx.Client:
        return httpx.Client(**_get_kwargs(self._settings))

    @cached_property
    def _auth_client(self) -> httpx.Client:
        return httpx.Client(**_get_auth_kwargs(self._settings))

    def close(self) -> None:
        self._client.close()
        self._auth_client.close()

    def __enter__(self) -> "GigaChatSyncClient":
        return self

    def __exit__(self, *exc: Any) -> None:
        self.close()

    def _update_token(self) -> None:
        if self._settings.credentials:
            self._access_token = post_auth.sync(
                self._auth_client,
                url=self._settings.auth_url,
                credentials=self._settings.credentials,
                scope=self._settings.scope,
            )
            logger.info("OAUTH UPDATE TOKEN")
        elif self._settings.user and self._settings.password:
            self._access_token = _build_access_token(
                post_token.sync(self._client, user=self._settings.user, password=self._settings.password)
            )
            logger.info("UPDATE TOKEN")

    def _decorator(self, call: Callable[..., T]) -> T:
        if self._use_auth:
            if self._check_validity_token():
                try:
                    return call()
                except AuthenticationError:
                    logger.warning("AUTHENTICATION ERROR")
                    self._reset_token()
            self._update_token()
        return call()

    def get_models(self) -> Models:
        """Возвращает массив объектов с данными доступных моделей"""
        return self._decorator(lambda: get_models.sync(self._client, access_token=self.token))

    def get_model(self, model: str) -> Model:
        """Возвращает объект с описанием указанной модели"""
        return self._decorator(lambda: get_model.sync(self._client, model=model, access_token=self.token))

    def chat(self, payload: Union[Chat, Dict[str, Any]]) -> ChatCompletion:
        """Возвращает ответ модели с учетом переданных сообщений"""
        chat = _parse_chat(payload, model=self._settings.model)
        return self._decorator(lambda: post_chat.sync(self._client, chat=chat, access_token=self.token))

    def stream(self, payload: Union[Chat, Dict[str, Any]]) -> Iterator[ChatCompletionChunk]:
        """Возвращает ответ модели с учетом переданных сообщений"""
        chat = _parse_chat(payload, model=self._settings.model)

        if self._use_auth:
            if self._check_validity_token():
                try:
                    for chunk in stream_chat.sync(self._client, chat=chat, access_token=self.token):
                        yield chunk
                    return
                except AuthenticationError:
                    logger.warning("AUTHENTICATION ERROR")
                    self._reset_token()
            self._update_token()

        for chunk in stream_chat.sync(self._client, chat=chat, access_token=self.token):
            yield chunk


class GigaChatAsyncClient(_BaseClient):
    """Асинхронный клиент GigaChat"""

    @cached_property
    def _aclient(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(**_get_kwargs(self._settings))

    @cached_property
    def _auth_aclient(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(**_get_auth_kwargs(self._settings))

    async def aclose(self) -> None:
        await self._aclient.aclose()
        await self._auth_aclient.aclose()

    async def __aenter__(self) -> "GigaChatAsyncClient":
        return self

    async def __aexit__(self, *exc: Any) -> None:
        await self.aclose()

    async def _aupdate_token(self) -> None:
        if self._settings.credentials:
            self._access_token = await post_auth.asyncio(
                self._auth_aclient,
                url=self._settings.auth_url,
                credentials=self._settings.credentials,
                scope=self._settings.scope,
            )
            logger.info("OAUTH UPDATE TOKEN")
        elif self._settings.user and self._settings.password:
            self._access_token = _build_access_token(
                await post_token.asyncio(self._aclient, user=self._settings.user, password=self._settings.password)
            )
            logger.info("UPDATE TOKEN")

    async def _adecorator(self, acall: Callable[..., Awaitable[T]]) -> T:
        if self._use_auth:
            if self._check_validity_token():
                try:
                    return await acall()
                except AuthenticationError:
                    logger.warning("AUTHENTICATION ERROR")
                    self._reset_token()
            await self._aupdate_token()
        return await acall()

    async def aget_models(self) -> Models:
        """Возвращает массив объектов с данными доступных моделей"""

        async def _acall() -> Models:
            return await get_models.asyncio(self._aclient, access_token=self.token)

        return await self._adecorator(_acall)

    async def aget_model(self, model: str) -> Model:
        """Возвращает объект с описанием указанной модели"""

        async def _acall() -> Model:
            return await get_model.asyncio(self._aclient, model=model, access_token=self.token)

        return await self._adecorator(_acall)

    async def achat(self, payload: Union[Chat, Dict[str, Any]]) -> ChatCompletion:
        """Возвращает ответ модели с учетом переданных сообщений"""
        chat = _parse_chat(payload, model=self._settings.model)

        async def _acall() -> ChatCompletion:
            return await post_chat.asyncio(self._aclient, chat=chat, access_token=self.token)

        return await self._adecorator(_acall)

    async def astream(self, payload: Union[Chat, Dict[str, Any]]) -> AsyncIterator[ChatCompletionChunk]:
        """Возвращает ответ модели с учетом переданных сообщений"""
        chat = _parse_chat(payload, model=self._settings.model)

        if self._use_auth:
            if self._check_validity_token():
                try:
                    async for chunk in stream_chat.asyncio(self._aclient, chat=chat, access_token=self.token):
                        yield chunk
                    return
                except AuthenticationError:
                    logger.warning("AUTHENTICATION ERROR")
                    self._reset_token()
            await self._aupdate_token()

        async for chunk in stream_chat.asyncio(self._aclient, chat=chat, access_token=self.token):
            yield chunk



from gigachat.pydantic_v1 import BaseModel, Field


class Model(BaseModel):
    """Описание модели"""

    id_: str = Field(alias="id")
    """Название модели"""
    object_: str = Field(alias="object")
    """Тип сущности в ответе, например, модель"""
    owned_by: str
    """Владелец модели"""
